from typing import Tuple


################################################################################
# TypeCheckError
#
# This exception class stores a classnotation typecheck failure. It stores all
# of the information needed in order to direct a user to the location in their
# dataclass where the error was found.
#
# TODO: We might also want to surface the data which was being processed as
#       well as in our real world testing we found that it was difficult to
#       find that data if it were being streamed into the program.
################################################################################
class TypeCheckError(Exception):
    message: str
    stack: Tuple[str, ...]

    def __init__(self, message: str, stack: Tuple[str, ...]):
        self.message = message
        self.stack = stack

    def __str__(self) -> str:
        return "PATH: .{stack} MESSAGE: {message}".format(
            stack=".".join(self.stack),
            message=self.message,
        )
