from classnotation.json_types import JSONData
from typing import Dict, Optional, Any


################################################################################
# UndefinedFields
#
# Undefined fileds is a wrapper class for storing all of the undefined fields
# within datum that do not appear within a dataclass.
#
# If a dataclass is allowed to support undefined fields then add a new variable
# to the dataclass with the type UndefinedFields.
#
#     @dataclass()
#     class MyClass:
#         myvar: int
#         _undefined_fields: UndefinedFields
#
# The field does not need to contain a default value, but may if it is more
# convienient to the user to add one
#
#     @dataclass()
#     class MyClass:
#         myvar: int
#         _undefined_fields: UndefinedFields = UndefinedFields()
#
################################################################################
class UndefinedFields():
    fields: Dict[str, JSONData]

    def __init__(self, fields: Optional[Dict[str, JSONData]] = None) -> None:
        if fields is None:
            self.fields = {}
        else:
            self.fields = fields

    def __eq__(self, other: Any) -> bool:
        if type(other) is not UndefinedFields:
            return False

        return self.fields == other.fields

    def __repr__(self) -> str:
        return f"UndefinedFields({self.fields})"
